from melusine.nlp_tools.tokenizer import Tokenizer
import numpy as np
from collections import Counter
from sklearn.base import BaseEstimator, ClassifierMixin
from keras.utils import np_utils
from keras.models import model_from_json
from keras.preprocessing.sequence import pad_sequences
from keras.optimizers import Adam


class NeuralModel(BaseEstimator, ClassifierMixin):
    """Generic class for  neural models.

    It is compatible with scikit-learn API (i.e. contains fit, transform
    methods).

    Parameters
    ----------
    neural_architecture_function : function,
        Function which returns a Model instance from Keras.

    pretrained_embedding : np.array,
        Pretrained embedding matrix.

    text_input_column : str,
        Input text column to consider for the model.

    meta_input_list : list, optional
        List of the names of the columns containing the metadata.
        If empty list or None the model is used without metadata
        Default value, ['extension', 'dayofweek', 'hour', 'min'].

    vocab_size : int, optional
        Size of vocabulary for neurol network model.
        Default value, 25000.

    seq_size : int, optional
        Maximum size of input for neural model.
        Default value, 100.

    loss : str, optional
        Loss function for training.
        Default value, 'categorical_crossentropy'.

    batch_size : int, optional
        Size of batches for the training of the neural network model.
        Default value, 4096.

    n_epochs : int, optional
        Number of epochs for the training of the neural network model.
        Default value, 15.

    Attributes
    ----------
    architecture_function, pretrained_embedding, text_input_column,
    meta_input_list, vocab_size, seq_size, loss, batch_size, n_epochs,

    model : Model instance from Keras,

    tokenizer : Tokenizer instance from Melusine,

    embedding_matrix : np.array,
        Embedding matrix used as input for the neural network model.

    Examples
    --------
    >>> from melusine.models.train import NeuralModel
    >>> from melusine.models.neural_architectures import cnn_model
    >>> from melusine.nlp_tools.embedding import Embedding
    >>> pretrained_embedding = Embedding.load()
    >>> list_meta = ['extension', 'dayofweek', 'hour']
    >>> nn_model = NeuralModel(cnn_model, pretrained_embedding, list_meta)
    >>> nn_model.fit(X_train, y_train)
    >>> y_res = nn_model.predict(X_test)

    """

    def __init__(self,
                pretrained_embedding=None,
                 architecture_function=None,
                 text_input_column='clean_text',
                 meta_input_list=['extension', 'dayofweek', 'hour', 'min'],
                 vocab_size=25000,
                 seq_size=100,
                 embedding_dim=200,
                 loss='categorical_crossentropy',
                 batch_size=4096,
                 n_epochs=15,
                 **kwargs):
        self.architecture_function = architecture_function
        self.pretrained_embedding = pretrained_embedding
        self.tokenizer = Tokenizer(input_column = text_input_column)
        self.text_input_column = text_input_column
        self.meta_input_list = meta_input_list
        self.vocab_size = vocab_size
        self.seq_size = seq_size
        self.embedding_dim = embedding_dim
        self.loss = loss
        self.batch_size = batch_size
        self.n_epochs = n_epochs

    def save_nn_model(self, filepath):
        """Save model to pickle, json and save weights to .h5."""
        json_model = self.model.to_json()
        open(filepath+".json", 'w').write(json_model)
        self.model.save_weights(filepath+"_model_weights.h5", overwrite=True)
        pass

    def load_nn_model(self, filepath):
        """Save model from json and load weights from .h5."""
        model = model_from_json(open(filepath+".json").read())
        model.load_weights(filepath+"_model_weights.h5")
        model.compile(optimizer=Adam(),
                      loss='categorical_crossentropy',
                      metrics=['accuracy'])
        self.model = model
        pass

    def __getstate__(self):
        """Method called before serialization for a specific treatment to save
        model weight and structure instead of standard serialization."""
        dict_attr = dict(self.__dict__)
        if 'model' in dict_attr:
            del dict_attr["model"]
            del dict_attr["embedding_matrix"]
            del dict_attr["pretrained_embedding"]
        return dict_attr

    def __setstate__(self, dict_attr):
        """Method called before loading class for a specific treatment to load
        model weight and structure instead of standard serialization."""
        self.__dict__ = dict_attr

    def fit(self, X, y, **kwargs):
        """Fit the neural network model on X and y.
        If meta_input list is empty list or None the model is used
        without metadata.

        Compatible with scikit-learn API.

        Parameters
        ----------
        X : pd.DataFrame

        y : pd.Series

        Returns
        -------
        self : object
            Returns the instance
        """

        X = self.tokenizer.transform(X)
        if self.pretrained_embedding:
            self._get_embedding_matrix()
        else:
            self._create_vocabulary_from_tokens(X)
            self._generate_random_embedding_matrix()
        X_seq = self._prepare_sequences(X)
        X_meta, nb_meta_features = self._get_meta(X)
        y_categorical = np_utils.to_categorical(y)
        nb_labels = len(np.unique(y))

        self.model = self.architecture_function(
            embedding_matrix_init=self.embedding_matrix,
            ntargets=nb_labels,
            seq_max=self.seq_size,
            nb_meta=nb_meta_features,
            loss=self.loss)

        if nb_meta_features == 0:
            X_input = X_seq
        else:
            X_input = [X_seq, X_meta]

        self.model.fit(X_input,
                       y_categorical,
                       batch_size=self.batch_size,
                       epochs=self.n_epochs,
                       **kwargs)
        pass

    def predict(self, X, **kwargs):
        """Returns the class predicted.

        Parameters
        ----------
        X : pd.DataFrame

        Returns
        -------
        int
        """
        return np.argmax(self.predict_proba(X, **kwargs), axis=1)

    def predict_proba(self, X, **kwargs):
        """Returns the probabilities associated to each classes.
        If meta_input list is empty list or None the model is used
        without metadata.

        Parameters
        ----------
        X : pd.DataFrame

        Returns
        -------
        np.array
        """
        X = self.tokenizer.transform(X)
        X_seq = self._prepare_sequences(X)
        X_meta, nb_meta_features = self._get_meta(X)
        if nb_meta_features == 0:
            X_input = X_seq
        else:
            X_input = [X_seq, X_meta]
        return self.model.predict(X_input, **kwargs)

    def _create_vocabulary_from_tokens(self, X):
        """Create a word indexes dictionary from tokens."""
        tokens = X['tokens']
        c = Counter([token for text in X.tokens for token in text])
        self.vocabulary = [t[0] for t in c.most_common(self.vocab_size)]
        pass

    def _get_embedding_matrix(self):
        """Prepares the embedding matrix to be used as an input for
        the neural network model.
        The vocabulary of the NN is those of the pretrained embedding
        """
        pretrained_embedding = self.pretrained_embedding
        self.vocabulary = pretrained_embedding.embedding.wv.index2word
        vocab_size = len(self.vocabulary)
        vector_dim = pretrained_embedding.embedding.vector_size
        embedding_matrix = np.zeros((vocab_size + 2, vector_dim))
        for index, word in enumerate(self.vocabulary):
            if word not in ['PAD', 'UNK']:
                embedding_matrix[index + 2, :] = pretrained_embedding.embedding.wv.get_vector(word)
        embedding_matrix[1, :] = np.mean(embedding_matrix, axis=0)

        self.vocabulary.insert(0, 'PAD')
        self.vocabulary.insert(1, 'UNK')
        self.embedding_matrix = embedding_matrix
        pass

    def _generate_random_embedding_matrix(self):
        """Prepares the embedding matrix to be used as an input for
        the neural network model.
        The vocabulary of the NN is those of the pretrained embedding
        """
        vocab_size = len(self.vocabulary)
        vector_dim = self.embedding_dim
        embedding_matrix = np.random.uniform(low=-1,high=1,size=(vocab_size + 2, vector_dim))
        embedding_matrix[0:2, :] = np.zeros((2, vector_dim))
        self.vocabulary.insert(0, 'PAD')
        self.vocabulary.insert(1, 'UNK')
        self.embedding_matrix = embedding_matrix
        pass

    def tokens_to_indices(self, tokens):
        """
        Input : list of tokens ["ma", "carte_verte", ...]
        Output : list of indices [46, 359, ...]
        """
        vocabulary_dict = {word: i for i, word in enumerate(self.vocabulary)}
        return [vocabulary_dict.get(token, 1) for token in tokens]

    def _prepare_sequences(self, X):
        """Prepares the sequence to be used as input for the neural network
        model.
        The input column must be an already tokenized text : tokens
        The tokens must have been optained using the same tokenizer than the
        one used for the pre-trained embedding."""
        seqs = X['tokens'].apply(self.tokens_to_indices)
        X_seq = pad_sequences(seqs, maxlen=self.seq_size)
        return X_seq

    def _get_meta(self, X):
        """Returns as a pd.DataFrame the metadata from X given the list_meta
        defined, and returns the number of columns. If meta_input_list is
        empty list or None, meta_input_list is returned as 0."""
        if self.meta_input_list is None or self.meta_input_list == []:
            X_meta = None
            nb_meta_features = 0
        else:
            meta_input_list = self.meta_input_list
            meta_input_list = [col+'__' for col in meta_input_list]
            columns_list = list(X.columns)
            meta_columns_list = [col for col in columns_list
                                 if col.startswith(tuple(meta_input_list))]
            X_meta = X[meta_columns_list]
            nb_meta_features = len(meta_columns_list)

        return X_meta, nb_meta_features



import sys
import logging
import gensim
import pickle
import re
from melusine.utils.streamer import Streamer
from melusine.config.config import ConfigJsonReader

conf_reader = ConfigJsonReader()
config = conf_reader.get_config_file()
common_terms = config["words_list"]["stopwords"] + config["words_list"]["names"]

regex_tokenize_with_punctuations = r"(.*?[\s'])"
tokenize_without_punctuations = r"(.*?)[\s']"
regex_process = "\w+(?:[\?\-\'\"_]\w+)*"
regex_split_parts = r"(.*?[;.,?!])"


log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s \
                              - %(message)s', datefmt='%d/%m %I:%M')


def phraser_on_body(row, phraser):
    """Applies phraser on cleaned body.

    To be used with methods such as: `apply(func, axis=1)` or
    `apply_by_multiprocessing(func, axis=1, **kwargs)`.

    Parameters
    ----------
    row : row of pd.Dataframe

    phraser : Phraser instance,

    Returns
    -------
    pd.Series

    Examples
    --------
        >>> import pandas as pd
        >>> data = pd.read_pickle('./tutorial/data/emails_anonymized.pickle')
        >>> from melusine.nlp_tools.phraser import phraser_on_body
        >>> from melusine.nlp_tools.phraser import Phraser
        >>> # data contains a 'clean_body' column
        >>> phraser = Phraser(columns='clean_body').load(filepath)
        >>> data.apply(phraser_on_body, axis=1)  # apply to all samples

    """
    clean_body = phraser_on_text(row["clean_body"], phraser)

    return clean_body


def phraser_on_header(row, phraser):
    """Applies phraser on cleaned header.

    To be used with methods such as: `apply(func, axis=1)` or
    `apply_by_multiprocessing(func, axis=1, **kwargs)`.

    Parameters
    ----------
    row : row of pd.Dataframe

    phraser : Phraser instance,

    Returns
    -------
    pd.Series

    Examples
    --------
        >>> import pandas as pd
        >>> data = pd.read_pickle('./tutorial/data/emails_anonymized.pickle')
        >>> from melusine.nlp_tools.phraser import phraser_on_header
        >>> from melusine.nlp_tools.phraser import Phraser
        >>> # data contains a 'clean_header' column
        >>> phraser = Phraser(columns='clean_header').load(filepath)
        >>> data.apply(phraser_on_header, axis=1)  # apply to all samples

    """
    clean_header = phraser_on_text(row["clean_header"], phraser)

    return clean_header


def phraser_on_text(text, phraser):
    """Returns text with phrased words.

    Parameters
    ----------
    text : str,

    phraser : Phraser instance,

    Returns
    -------
    str

    """
    if not re.search(pattern=r"\W*\b\w+\b\W*", string=text):
        return text
    pre_typos_list, words_list, separators_list = _split_typos_words_separators(text)
    phrased_words_list = phraser.phraser[words_list]
    phrased_text = _rebuild_phrased_text_with_punctuation(pre_typos_list,
                                                          words_list,
                                                          separators_list,
                                                          phrased_words_list)

    return phrased_text


def _rebuild_phrased_text_with_punctuation(pre_typos_list,
                                           words_list,
                                           separators_list,
                                           phrased_words_list):
    """Rebuilds the initial text with phrased words."""
    i = 0
    for pre_typo, word, separator in zip(pre_typos_list,
                                         words_list,
                                         separators_list):
        phrased_word = re.sub("\W", "", phrased_words_list[i])
        word = re.sub("\W", "", word)
        if len(phrased_word) > len(word):
            if _check_last_word_phrased(phrased_word, word):
                phrased_words_list[i] = pre_typo + phrased_word + separator
                i += 1
        else:
            phrased_words_list[i] = pre_typo + phrased_word + separator
            i += 1

    return "".join(phrased_words_list)


def _check_last_word_phrased(phrased_word, word):
    """Check if a word is the last word of a phrased word."""
    words_list = phrased_word.split("_")
    last_word = words_list[-1]

    return word == last_word


def _split_typos_words_separators(text, pattern=r"(\W*)\b(\w+)\b(\W*)"):
    """Split text according to typos."""
    tuple_word_separator_list = re.findall(pattern, text, flags=re.M | re.I)
    pre_typos_list, words_list, separators_list = zip(*tuple_word_separator_list)

    return pre_typos_list, words_list, separators_list


class Phraser():
    """Class to train a phraser.

    Parameters
    ----------
    input_column : str,
        Input text column to consider for the phraser.

    common_terms : list of integers, optional
        List of stopwords.
        Default value, list of stopwords from nltk.

    threshold : int, optional
        Threshold to select colocations.
        Default value, 350.

    min_count : int, optional
        Minimum count of word to be selected as colocation.
        Default value, 200.

    Attributes
    ----------
    common_terms, threshold, min_count,

    stream : Streamer object,
        Builds a stream a tokens from a pd.Dataframe to train the embeddings.

    phraser : Phraser object from Gensim

    Examples
    --------
    >>> from melusine.nlp_tools.phraser import Phraser
    >>> phraser = Phraser()
    >>> phraser.train(X)
    >>> phraser.save(filepath)
    >>> phraser = phraser().load(filepath)

    """

    def __init__(self,
                 input_column='clean_body',
                 common_terms=common_terms,
                 threshold=350,
                 min_count=200):
        self.logger = logging.getLogger('NLUtils.Phraser')
        self.logger.debug('creating a Phraser instance')
        self.common_terms = common_terms
        self.threshold = threshold
        self.min_count = min_count
        self.input_column = input_column
        self.streamer = Streamer(column=self.input_column, stop_removal=False)
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.INFO)
        ch.setFormatter(formatter)
        self.logger.addHandler(ch)

    def __getstate__(self):
        """should return a dict of attributes that will be pickled
        To override the default pickling behavior and
        avoid the pickling of the logger
        """
        d = self.__dict__.copy()
        if 'logger' in d:
            d['logger'] = d['logger'].name
        return d

    def __setstate__(self, d):
        """To override the default pickling behavior and
        avoid the pickling of the logger"""
        if 'logger' in d:
            d['logger'] = logging.getLogger(d['logger'])
        self.__dict__.update(d)

    def save(self, filepath):
        """Method to save Phraser object"""
        with open(filepath, 'wb') as f:
            pickle.dump(self.phraser, f)

    def load(self, filepath):
        """Method to load Phraser object"""
        with open(filepath, 'rb') as f:
            self.phraser = pickle.load(f)
        return self

    def train(self, X):
        """Train phraser.

        Parameters
        ----------
        X : pd.Dataframe

        Returns
        -------
        self : object
            Returns the instance
        """
        self.logger.info('Start training for colocation detector')
        self.streamer.to_stream(X)
        phrases = gensim.models.Phrases(self.streamer.stream,
                                        common_terms=self.common_terms,
                                        threshold=self.threshold,
                                        min_count=self.min_count)
        self.phraser = gensim.models.phrases.Phraser(phrases)
        self.logger.info('Done.')
        pass



"""
Useful class to define its own transformer using specific functions
in a specific order to apply along a row of DataFrame (axis=1).

It is compatible with scikit-learn API (i.e. contains fit, transform methods).
"""

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from melusine.utils.multiprocessing import apply_by_multiprocessing


def __check_function_type(func):
    """Check if it is a function-like object."""
    if not callable(func):
        raise TypeError("First item of the tuple (func, args, cols) must be a function-like \
object not a {} object".format(type(func)))
    else:
        return func


def __check_args_type(args):
    """Check if it is a tuple-like object."""
    if args is None or args == ():
        return None
    elif isinstance(args, int) or isinstance(args, str) or isinstance(args, list):
        # manage the case of 1 element in tuple (example: args=(4))
        return (args, )
    elif not isinstance(args, tuple):
        raise TypeError("Second item of the tuple (func, args, cols) must be tuple-like \
object not a {} object".format(type(args)))
    else:
        return args


def __check_colnames_type(cols):
    """Check if it is a list-like object."""
    if cols is None or cols == []:
        return None
    elif not isinstance(cols, list):
        raise TypeError("Third item of the tuple (func, args, cols) must be list-like \
object not a {} object".format(type(cols)))
    else:
        return cols


def _check_tuple(func, args=None, cols=None):
    """Complete checking of each element for the 'function_scheduler'
    parameter."""
    # check types of each parameters
    func = __check_function_type(func)
    args = __check_args_type(args)
    cols = __check_colnames_type(cols)

    return (func, args, cols)


class TransformerScheduler(BaseEstimator, TransformerMixin):
    """
    This class aims to provide a good way to define its own transformer.
    It takes a list of function defined in a specific order to apply along a
    row of DataFrame (axis=1).
    Transformer returned is compatible with scikit-learn API
    (i.e. contains fit, transform methods).

    Parameters
    ----------
    functions_scheduler : list of tuples, (function, tuple, list)
        List of function to be applied in a specific order.
        Each element of the list has to be defined as follow:
        (`function`, `argument(s) used by the function (optional)`, `colname(s)
        returned (optional)`)

    mode : str {'apply', 'apply_by_multiprocessing'}, optional
        Define mode to apply function along a row axis (axis=1).
        Default value, 'apply'.
        If set to 'apply_by_multiprocessing', it uses multiprocessing tool
        to parallelize computation.

    n_jobs : int, optional
        Number of cores used for computation. Default value, 1.

    progress_bar : boolean, optional
        Whether to print a progress bar from tqdm package. Default value, True.
        Works only when mode is set to 'apply_by_multiprocessing'.

    copy : boolean, optional
        Make a copy of DataFrame. Default value, True.

    verbose : int, optional
        Verosity mode, print loggers. Default value, 0.

    Attributes
    ----------
    function_scheduler, mode, n_jobs, progress_bar

    Examples
    --------
    >>> from melusine.utils.transformer_scheduler import TransformerScheduler

    >>> MelusineTransformer = TransformerScheduler(
    >>>     functions_scheduler=[
    >>>         (my_function_1, (argument1, argument2), ['return_col_A']),
    >>>         (my_function_2, None, ['return_col_B', 'return_col_C'])
    >>>         (my_function_3, (), ['return_col_D'])
    >>>     ])

    """

    def __init__(self,
                 functions_scheduler,
                 mode='apply',
                 n_jobs=1,
                 progress_bar=True,
                 copy=True,
                 verbose=0):
        self.functions_scheduler = functions_scheduler
        self.mode = mode
        self.n_jobs = n_jobs
        self.progress_bar=True
        self.copy = copy
        self.verbose = verbose

        # check input parameters type
        for tuple_ in functions_scheduler:
            func, args, cols = _check_tuple(*tuple_)

    def fit(self, X, y=None):
        """Unused method. Defined only for compatibility with scikit-learn API.
        """
        return self

    def transform(self, X):
        """Apply functions defined in the `function_scheduler` parameter.

        Parameters
        ----------
        X : pandas.DataFrame,
            Data on which transformations are applied.

        Returns
        -------
        pandas.DataFrame
        """
        if self.copy:
            X_ = X.copy()
        else:
            X_ = X

        for tuple_ in self.functions_scheduler:
            func_, args_, cols_ = _check_tuple(*tuple_)
            # cols_ = cols_ or X_.columns

            if self.mode == 'apply':
                if cols_ is None:
                    X_ = X_.apply(func_, args=args_, axis=1)
                elif len(cols_) == 1:
                    X_[cols_[0]] = X_.apply(func_, args=args_, axis=1)
                else:
                    X_[cols_] = X_.apply(func_, args=args_, axis=1).apply(pd.Series)
            else:  # 'apply_by_multiprocessing'
                if cols_ is None:
                    X_ = apply_by_multiprocessing(df=X_,
                                                    func=func_,
                                                    args=args_,
                                                    axis=1,
                                                    workers=self.n_jobs,
                                                    progress_bar=self.progress_bar)
                elif len(cols_) == 1:
                    X_[cols_[0]] = apply_by_multiprocessing(df=X_,
                                                    func=func_,
                                                    args=args_,
                                                    axis=1,
                                                    workers=self.n_jobs,
                                                    progress_bar=self.progress_bar)
                else:
                    X_[cols_] = apply_by_multiprocessing(df=X_,
                                                    func=func_,
                                                    args=args_,
                                                    axis=1,
                                                    workers=self.n_jobs,
                                                    progress_bar=self.progress_bar).apply(pd.Series)
        return X_



import re
from melusine.config.config import ConfigJsonReader
from melusine.prepare_email.cleaning import remove_accents

conf_reader = ConfigJsonReader()
config = conf_reader.get_config_file()

newline_character = config['regex']['cleaning']['newline_character']
signature_token_threshold = config['regex']['mail_segmenting']['signature_token_threshold']

REGEX_TR_RE = config['regex']['manage_transfer_reply']
REGEX_SEG = config['regex']['mail_segmenting']

regex_begin_transfer = REGEX_TR_RE['begin_transfer']
regex_transfer_other = REGEX_TR_RE['transfer_other']
regex_extract_from = REGEX_TR_RE['extract_from']
regex_extract_to = REGEX_TR_RE['extract_to']
regex_extract_date = REGEX_TR_RE['extract_date']
regex_extract_header = REGEX_TR_RE['extract_header']
regex_answer_header = REGEX_TR_RE['answer_header']
regex_transfert_header = REGEX_TR_RE['transfer_header']

regex_tag = REGEX_SEG['tag']
regex_segmenting_dict = REGEX_SEG['segmenting_dict']
regex_segmenting_dict['RE/TR'] = [regex_begin_transfer,
                                  regex_transfer_other,
                                  regex_extract_from,
                                  regex_extract_to,
                                  regex_extract_date,
                                  regex_extract_header]

compiled_regex_segmenting_dict = {}
for tag, regex_list in regex_segmenting_dict.items():
    compiled_regex_segmenting_dict[tag] = [
        re.compile(regex.replace(" ", regex_tag), re.I)
        for regex in regex_list
    ]

regex_from1 = REGEX_SEG['meta_from1']
regex_from2 = REGEX_SEG['meta_from2']
regex_to = REGEX_SEG['meta_to']
regex_date1 = REGEX_SEG['meta_date1']
regex_date2 = REGEX_SEG['meta_date2']
regex_header = REGEX_SEG['meta_header']
regex_piece_jointe = REGEX_SEG['pattern_pj']

regex_exception_une_lettre_maj = REGEX_SEG['pattern_exception_une_lettre_maj']
regex_exception_Mr = REGEX_SEG['pattern_exception_Mr']
regex_exception_Dr = REGEX_SEG['pattern_exception_Dr']
regex_exception_Mme = REGEX_SEG['pattern_exception_Mme']
regex_exception = REGEX_SEG['pattern_exception']
regex_pattern_exceptions = (regex_exception_une_lettre_maj
                            + regex_exception_Mr
                            + regex_exception_Dr
                            + regex_exception_Mme
                            + regex_exception)

regex_sep_doubles_points_virgules_espace = REGEX_SEG['pattern_sep_doubles_points_virgules_espace']
regex_pattern_separteurs_evidents = REGEX_SEG['pattern_separteurs_evidents']
regex_pattern_beginning = REGEX_SEG['pattern_beginning']
regex_pattern_end = REGEX_SEG['pattern_end']
regex_pattern = (regex_pattern_beginning
                 + regex_pattern_separteurs_evidents
                 + regex_sep_doubles_points_virgules_espace
                 + regex_pattern_exceptions
                 + regex_pattern_end)

compiled_regex_typo = re.compile(REGEX_SEG['tag_typo'], re.I)
regex_tag_subsentence = REGEX_SEG['tag_subsentence']
regex_split_message_to_sentences_list = REGEX_SEG['split_message_to_sentences_list']


def structure_email(row):
    """ 1. Splits parts of each messages in historic and tags them.
    For example a tag can be hello, body, greetings etc
    2. Extracts the meta informations of each messages

    To be used with methods such as: `apply(func, axis=1)` or
    `apply_by_multiprocessing(func, axis=1, **kwargs)`.

    Parameters
    ----------
    row : row of pd.Dataframe, apply on column ['structured_historic']

    Returns
    -------
    list of dicts : one dict per message

    Examples
    --------
        >>> import pandas as pd
        >>> from melusine.prepare_email.build_historic import build_historic
        >>> data = pd.read_pickle('./tutorial/data/emails_anonymized.pickle')
        >>> data['structured_historic'] = data.apply(build_historic, axis=1)
        >>> # data contains column ['structured_historic']

        >>> from melusine.prepare_email.mail_segmenting import structure_email
        >>> structure_email(data.iloc[0])  # apply for 1 sample
        >>> data.apply(structure_email, axis=1)  # apply to all samples

    """
    structured_body = []
    for message in row['structured_historic']:
        structured_message = structure_message(message)
        if len(structured_message["structured_text"]["text"]) == 0:
            if structured_message["structured_text"]["header"] is None:
                continue
        structured_body.append(structured_message)

    return structured_body


def structure_message(message):
    """ Splits parts of a message and tags them.
    For example a tag can be hello, body, greetings etc
    Extracts the meta informations of the message

    Parameters
    ----------
    message : dict

    Returns
    -------
    dict

    Examples
    --------
    """
    meta = str(message.get("meta"))
    structured_meta, header = structure_meta(meta)
    text = str(message.get("text"))
    tagged_parts_list = tag_parts_message(text)
    structured_message = _tuples_to_dict(structured_meta,
                                         header,
                                         tagged_parts_list)

    return structured_message


def structure_meta(meta):
    """ Extract meta informations (date, from, to, header) from string meta

    Parameters
    ----------
    meta : str

    Returns
    -------
    tuple(dict, string)

    Examples
    --------
    """
    structured_meta = {}
    structured_meta['date'] = _find_date(meta)
    structured_meta['from'] = _find_from(meta)
    structured_meta['to'] = _find_meta(regex_to, meta)
    header = _find_meta(regex_header, meta)

    return structured_meta, header


def _find_date(message):
    """ Match pattern regex with a given message """
    group = _find_meta(regex_date1, message)
    if group is None:
        group = _find_meta(regex_date2, message)

    return group


def _find_from(message):
    """ Match pattern regex with a given message """
    group = _find_meta(regex_from1, message)
    if group is None:
        group = _find_meta(regex_from2, message)

    return group


def _find_meta(regex, message):
    """ Match pattern regex with a given message """
    groups = re.findall(regex, message)
    if len(groups) < 1:
        return None
    else:
        return groups[0]


def tag_parts_message(text):
    """ Splits message into sentences, tags them and merges two sentences in a
    row having the same tag.

    Parameters
    ----------
    text : str,


    Returns
    -------
    list of tuples

    Examples
    --------
    """
    sentence_list = split_message_to_sentences(text)
    tagged_sentence_list = []
    for sentence in sentence_list:
        tagged_sentence = tag_sentence(sentence)
        tagged_sentence_list.extend(tagged_sentence)
    tagged_parts_list = _merge_parts(tagged_sentence_list)
    tagged_parts_list = _remove_empty_parts(tagged_parts_list)
    tagged_parts_list = _update_typo_parts(tagged_parts_list)
    tagged_parts_list = _remove_typo_parts(tagged_parts_list)

    return tagged_parts_list


def split_message_to_sentences(text, sep_=r"(.*?[;.,?!])"):
    """ Split each sentences in a text """
    regex1 = regex_split_message_to_sentences_list[0]
    regex2 = regex_split_message_to_sentences_list[1]
    regex3 = regex_split_message_to_sentences_list[2]
    regex4 = regex_split_message_to_sentences_list[3]
    text = text.strip(regex1).lstrip(regex2)
    text = re.sub(regex3, regex4, text)      # remove double punctuation
    sentence_list = re.findall(regex_pattern, text, flags=re.M)
    sentence_list = [r for s in sentence_list
                     for r in re.split(regex_piece_jointe, s) if r]

    return sentence_list


def tag_sentence(sentence, default='BODY'):
    """ Tag a sentence.
    If the sentence cannot be tagged it will tag the subsentences

    Parameters
    ----------
    sentence : str,


    Returns
    -------
    list of tuples : sentence, tag

    Examples
    --------
    """
    tagged_sentence, tagged = tag(sentence)
    if tagged:
        return tagged_sentence
    else:
        return _tag_subsentence(sentence)


def _tag_subsentence(sentence, default='BODY'):
    """ Tags the subsentences in a sentence.
    If the subsentences cannot be tagged it will return the whole sentence with
    a default tag.

    Parameters
    ----------
    sentence : str,


    Returns
    -------
    list of tuples : sentence, tag

    Examples
    --------
    """
    subsentence_list = re.findall(regex_tag_subsentence, sentence, flags=re.M)
    tagged_subsentence_list = []
    any_sub_catch = False
    for subsentence in subsentence_list:
        tagged_subsentence, subcatch = tag(subsentence)
        if subcatch:
            tagged_subsentence_list.extend(tagged_subsentence)
            any_sub_catch = True
        else:
            tagged_subsentence_list.append((subsentence, default))
    if any_sub_catch:
        return tagged_subsentence_list
    else:
        return [(sentence, default)]


def tag(string):
    """ Tags a string.

    Parameters
    ----------
    string : str,


    Returns
    -------
    tuples : list of tuples and boolean

    Examples
    --------
    """
    regex_parts = compiled_regex_segmenting_dict.items()
    sentence_with_no_accent = remove_accents(string)
    for k, compiled_regex_list in regex_parts:
        for compiled_regex in compiled_regex_list:
            if compiled_regex.search(sentence_with_no_accent):
                return [(string, k)], True

    return string, False


def _merge_parts(list_de_tuple_parts_id):
    """ Merge two consecutives strings with the same tag """
    if len(list_de_tuple_parts_id) <= 1:
        return list_de_tuple_parts_id
    i = 0
    j = 1
    sentences, tags = zip(*list_de_tuple_parts_id)
    tags = list(tags)
    sentences = list(sentences)
    while j < len(list_de_tuple_parts_id):
        if tags[i] == tags[j]:
            sentences[i] = " ".join((sentences[i], sentences[j]))
            j += 1
        else:
            i += 1
            tags[i] = tags[j]
            sentences[i] = sentences[j]
            j += 1
    list_de_tuples_merged = list(zip(sentences[:i+1], tags[:i+1]))

    return list_de_tuples_merged


def _remove_empty_parts(tagged_parts_list):
    """ Remove all the empty parts in the list of tagged parts """
    tagged_parts_list = [part for part in tagged_parts_list
                         if len(part[0]) > 0]

    return tagged_parts_list


def _update_typo_parts(tagged_parts_list):
    """ Update the tagging for all the typo parts in the list of
    tagged parts """
    tagged_parts_list = [_update_typo_part(part_tag_tuple)
                         for part_tag_tuple in tagged_parts_list]

    return tagged_parts_list


def _update_typo_part(part_tag_tuple):
    part, tag = part_tag_tuple
    if __is_typo(part):
        part_tag_tuple = part, "TYPO"

    return part_tag_tuple


def __is_typo(part, compiled_regex_typo=compiled_regex_typo):
    """ Check if a string is typo """
    return compiled_regex_typo.search(part)


def _remove_typo_parts(tagged_parts_list):
    """  """
    tagged_parts_list = [part_tag_tuple
                         for part_tag_tuple in tagged_parts_list
                         if part_tag_tuple[1] != "TYPO"]

    return tagged_parts_list


def _tuples_to_dict(meta, header, tagged_parts):
    """ Convert a dictionnary and list of tuples into dictionnary """
    structured_message = {}
    structured_message["meta"] = meta
    structured_message["structured_text"] = {}
    structured_message["structured_text"]["header"] = header
    structured_text = []
    for part, tag in tagged_parts:
        dict_message = {}
        dict_message["part"] = part
        dict_message["tags"] = tag
        structured_text.append(dict_message)
    structured_message["structured_text"]["text"] = structured_text

    return structured_message


def tag_signature(row, token_threshold = signature_token_threshold):
    """
    Function to be called after the mail_segmenting function as it requires a "structured_body" column.
    This function detects parts of a message that qualify as "signature".
    Exemples of parts qualifying as signature are sender name, company name, phone number, etc.

    The methodology to detect a signature is the following:
    - Look for a THANKS or GREETINGS part indicating that the message is approaching the end
    - Check the length of the following message parts currently tagged as "BODY"
    - (The maximum number of words is specified through the variable "signature_token_threshold")
    - If ALL the "ending parts" contain few words => tag them as "SIGNATURE" parts
    - Otherwise : cancel the signature tagging

    Parameters
    ----------
    row : pd.Series
        Row of an email DataFrame

    Returns
    -------
    structured_body : Updated structured body
    """

    # Get the part and tags in the email
    last_body_parts = row['structured_body'][0]["structured_text"]['text']

    # Get index of the first occurrence of a THANKS or GREETINGS part
    ending_part_index = next((i for i, x in enumerate(last_body_parts) if x['tags'] in ['GREETINGS', 'THANKS']), -1)

    # Detect parts that qualify as signature
    signature_indices = _detect_signature_parts(last_body_parts, ending_part_index, token_threshold)

    # Modify tag for parts that qualify as SIGNATURE
    for signature_part_index in signature_indices:
        last_body_parts[signature_part_index]['tags'] = "SIGNATURE"

    return row['structured_body']


def _detect_signature_parts(last_body_parts, part_index, token_threshold):
    """
    Check the length of the "BODY" parts at the end of a message.
    If all the ending "BODY" parts contains fewer words than the specified threshold, return the indices of the parts.

    Parameters
    ----------
    last_body_parts : list of dict
        Tag and part content for the different parts in a message.
    part_index : int
        Index of the part tagged as "GREETINGS" or "THANKS"
    token_threshold : int
        Maximum number of words/tokens in a sentence to qualify as a SIGNATURE sentence

    Returns
    -------
    list : indices of the parts to be tagged as SIGNATURE
    """
    signature_indices = []

    # If at least 1 THANKS / GREETINGS part
    if part_index == -1:
        return signature_indices

    # Loop on parts AFTER the THANKS / GREETINGS part
    for i, part_tag in enumerate(last_body_parts[part_index + 1:]):

        # Check that part is a BODY part (ignore other parts)
        if part_tag['tags'] == 'BODY':

            # Split text to sentences (Because identical consecutive parts have been previously merged)
            sentences = part_tag['part'].split(newline_character)

            # Count number of words/tokens in sentences
            for sentence in sentences:
                n_words = len(re.sub(r"[;.,:!?]", "", sentence).split())

                # If at one sentence or more contains more than 4 words, cancel signature tagging
                if n_words >= token_threshold:
                    return []

            # If part qualifies as SIGNATURE, store index
            signature_indices.append(part_index + 1 + i)

    return signature_indices



"""Top-level package for Melusine."""

__author__ = """Sacha Samama & Tom Stringer & Hugo Perrier"""
__email__ = 'ssamama@quantmetry.com, tstringer@quantmetry.com, hperrier@quantmetry.com'
__version__ = '1.8.7'



import logging
import re
from sklearn.base import BaseEstimator, TransformerMixin
from melusine.config.config import ConfigJsonReader

conf_reader = ConfigJsonReader()
config = conf_reader.get_config_file()

stopwords = config["words_list"]["stopwords"]
names_list = config["words_list"]["names"]
regex_tokenize = config["regex"]["tokenizer"]


class Tokenizer(BaseEstimator, TransformerMixin):
    """Class to train and apply tokenizer.

    Compatible with scikit-learn API (i.e. contains fit, transform methods).

    Parameters
    ----------
    input_column : str,
        Input text column to consider for the tokenizer.

    stopwords : list of strings, optional
        List of words to remove from list of tokens.
        Default value, list defined in conf file

    stop_removal : boolean, optional
        True if stopwords to be removed, else False.
        Default value, False.

    n_jobs : int, optional
        Number of cores used for computation.
        Default value, 20.

    Attributes
    ----------
    stopwords, stop_removal, n_jobs

    Examples
    --------
    >>> from melusine.nlp_tools.tokenizer import Tokenizer
    >>> tokenizer = Tokenizer()
    >>> X = tokenizer.fit_transform(X)
    >>> tokenizer.save(filepath)
    >>> tokenizer = Tokenizer().load(filepath)

    """

    def __init__(self,
                 input_column='clean_text',
                 stopwords=stopwords,
                 stop_removal=True,
                 n_jobs=20):
        self.input_column = input_column
        self.stopwords = set(stopwords)
        self.stop_removal = stop_removal
        self.names_list = set(names_list)
        self.n_jobs = n_jobs
        self.logger = logging.getLogger(__name__)

    def __getstate__(self):
        """should return a dict of attributes that will be pickled
        To override the default pickling behavior and
        avoid the pickling of the logger
        """
        d = self.__dict__.copy()
        d['n_jobs'] = 1
        if 'logger' in d:
            d['logger'] = d['logger'].name
        return d

    def __setstate__(self, d):
        """To override the default pickling behavior and
        avoid the pickling of the logger"""
        if 'logger' in d:
            d['logger'] = logging.getLogger(d['logger'])
        self.__dict__.update(d)

    def fit(self, X, y=None):
        """Unused method. Defined only for compatibility with scikit-learn API.
        """
        return self

    def transform(self, X):
        """Applies tokenize method on pd.Dataframe.

        Parameters
        ----------
        X : pandas.DataFrame,
            Data on which transformations are applied.

        Returns
        -------
        pandas.DataFrame
        """
        self.logger.debug('Start transform tokenizing')
        X['tokens'] = X[[self.input_column]].apply(self.tokenize, axis=1)
        X['tokens'] = X.apply(lambda x: x['tokens'][0], axis=1)
        self.logger.info('X shape : %s' % str(X.shape))
        self.logger.debug('Done.')
        return X

    def tokenize(self, row):
        """Returns list of tokens.

        Parameters
        ----------
        row : row of pd.Dataframe

        Returns
        -------
        pd.Series

        """
        text = row[self.input_column]
        tokens = self._tokenize(text)
        return [tokens]

    def _tokenize(self, text, pattern=regex_tokenize):
        """Returns list of tokens from text."""
        if isinstance(text, str):
            tokens = re.findall(pattern, text, re.M+re.DOTALL)
            tokens = self._remove_stopwords(tokens)
        else:
            tokens = []
        return tokens

    def _remove_stopwords(self, list):
        """ Removes stopwords from list if stop_removal parameter
        set to True."""
        if self.stop_removal:
            return [tok if tok not in self.names_list else "flag_name_" for tok in list if tok not in self.stopwords]
        else:
            return list



import re
from melusine.config import ConfigJsonReader

conf_reader = ConfigJsonReader()
config = conf_reader.get_config_file()
regex_transition_list = config['regex']['build_historic']['transition_list']


def build_historic(row):
    """Rebuilds and structures historic of emails from the whole contents.
    Function has to be applied with `apply` method of a DataFrame along an
    axis=1.
    For each email of the historic, it segments the body into 2 different parts
    (2 keys of dict):

    {'text': extract raw text without metadata,
     'meta': get transition from the 'transition_list' defined in the conf.json
     }.


    Parameters
    ----------
    row : row,
        A pandas.DataFrame row object with 'body' column.

    Returns
    -------
    list

    Examples
    --------
        >>> import pandas as pd
        >>> data = pd.read_pickle('./tutorial/data/emails_anonymized.pickle')
        >>> # data contains a 'body' column

        >>> from melusine.prepare_email.build_historic import build_historic
        >>> build_historic(data.iloc[0])  # apply for 1 sample
        >>> data.apply(build_historic, axis=1)  # apply to all samples

    """
    email_body = row['body']
    index_messages, nb_messages = _get_index_transitions(email_body)
    structured_historic = [
        {'text': email_body[index_messages[i][1]:index_messages[i+1][0]],
         'meta': email_body[index_messages[i][0]:index_messages[i][1]]
         } for i in range(nb_messages)]
    structured_historic = __remove_empty_mails(structured_historic)
    return structured_historic


def _get_index_transitions(email_body):
    """Returns list of indexes defining the transitions between
    different messages in an email."""
    index = []
    for regex in regex_transition_list:
        for match in re.finditer(regex, email_body, flags=re.S):
            idx = (match.start(), match.end())
            index.append(idx)

    index = [(0, 0)] + index
    index = index + [(len(email_body), len(email_body))]
    index = sorted(list(set(index)))
    index = __filter_overlap(index)
    nb_parts = len(index) - 1

    return index, nb_parts


def __filter_overlap(index):
    """Filters indexes in list if they overlap."""
    if len(index) == 2:
        return index
    index_f = []
    i = 0
    j = i + 1
    while j < len(index):
        if index[i][1] > index[j][0]:
            index[i] = (min(index[i][0], index[j][1]),
                        max(index[i][0], index[j][1]))
            j += 1
        else:
            index_f += [index[i]]
            i = j
            j += 1
    index_f += [index[i]]

    return index_f[:i+1]


def is_only_typo(text):
   """check if the string contains any word character"""
   if not re.search(r"\w", text):
       return True
   else:
       return False


def __remove_empty_mails(structured_historic):
    """If an interval between two matches of transitions is empty (or typographic)
    then we remove this interval and we report the meta data to the previous email.
    Otherwise this empty interval would be considered as an email and the meta data
    will be lost"""
    purged_structured_historic = []
    meta_to_reinclude = None
    for text_meta in structured_historic:
        if meta_to_reinclude: # if the precedent piece had meta data to reinclude
            text_meta["meta"] = meta_to_reinclude + text_meta.get("meta")
            meta_to_reinclude = None
        text, meta = text_meta.get("text"), text_meta.get("meta")
        if not is_only_typo(text):
            purged_structured_historic.append(text_meta)
        else:
            if not is_only_typo(meta):
                meta_to_reinclude = meta
    if len(purged_structured_historic) == 0:
        purged_structured_historic = [{'text': '', 'meta': meta_to_reinclude}]
    return purged_structured_historic



import pandas as pd
import numpy as np
from melusine.prepare_email.manage_transfer_reply import (
    add_boolean_answer, add_boolean_transfer, check_mail_begin_by_transfer)


def test_add_boolean_answer():
    input_df = pd.DataFrame({
        'header': ["Bonjour, je suis disponible",
                   "RE: bonjour",
                   "re: bonjour",
                   "TR: bonjour",
                   "Fwd: bonjour",
                   "Re: bonjour",
                   np.nan,
                   ''],
    })

    output_df = pd.Series([False, True, True, False, False, True, False,
                           False])

    result = input_df.apply(add_boolean_answer, axis=1)
    pd.testing.assert_series_equal(result, output_df)


def test_add_boolean_transfer():
    input_df = pd.DataFrame({
        'header': ["Bonjour, je suis disponible",
                   "RE: bonjour",
                   "re: bonjour",
                   "TR: bonjour",
                   "Tr: bonjour",
                   "Fwd: bonjour",
                   "FW: bonjour",
                   "FWD: bonjour",
                   "Fw: bonjour",
                   np.nan,
                   ''],
    })

    output_df = pd.Series([False, False, False, True, True, True, True, True,
                           True, False, False])

    result = input_df.apply(add_boolean_transfer, axis=1)
    pd.testing.assert_series_equal(result, output_df)


def test_check_mail_begin_by_transfer():
    input_df = pd.DataFrame({
        'body': ["Bonjour, je suis disponible",
                 "--- Transféré par <xxxx@gmail.com> ---- Bonjour, je suis disponible",
                 "De : <xxxx@gmail.com> : salut, je suis disponible",
                 np.nan,
                 '']
    })

    output_df = pd.Series([False, True, True, False, False])

    result = input_df.apply(check_mail_begin_by_transfer, axis=1)
    pd.testing.assert_series_equal(result, output_df)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['pandas>=0.25.0',
                'scikit-learn>=0.19.0',
                'gensim>=3.3.0',
                'keras>=2.2.0',
                'tqdm>=4.34',
                'tensorflow>=1.10.0,<=1.13.1',
                'unidecode',
                'joblib'
                ]

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest', ]

setup(
    author="Sacha Samama, Tom Stringer, Antoine Simoulin, Benoit Lebreton",
    author_email='ssamama@quantmetry.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="Melusine is a high-level package for french emails preprocessing, classification and feature extraction, written in Python.",
    entry_points={},
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords='melusine',
    name='melusine',
    package_dir={
        'melusine': 'melusine',
        'melusine.config': 'melusine/config',
        'melusine.utils': 'melusine/utils',
        'melusine.nlp_tools': 'melusine/nlp_tools',
        'melusine.prepare_email': 'melusine/prepare_email',
        'melusine.summarizer': 'melusine/summarizer',
        'melusine.models': 'melusine/models',
        'melusine.data': 'melusine/data'
    },
    packages=['melusine', 'melusine.config', 'melusine.utils',
              'melusine.nlp_tools', 'melusine.prepare_email',
              'melusine.summarizer', 'melusine.models',
              'melusine.data'],
    data_files=[('config', ['melusine/config/conf.json']),
                ('data', ['melusine/data/emails.csv'])],
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/MAIF/melusine',
    version='1.8.7',
    zip_safe=False,
)



import pytest
import numpy as np
from melusine.prepare_email.cleaning import (
    remove_multiple_spaces_and_strip_text, remove_accents,
    flag_items, remove_transfer_answer_header)


@pytest.mark.parametrize("input_str, expected_str", [
    ("hello   world", "hello world"),
    ("\n   hello world    ", "hello world"),
    ("----- hello\tworld *****", "hello world"),
    ("hello-world", "hello-world"),
    ("hello - world", "hello world")
])
def test_remove_multiple_spaces_and_strip_text(input_str, expected_str):
    result = remove_multiple_spaces_and_strip_text(input_str)
    np.testing.assert_string_equal(result, expected_str)


def test_remove_accents():
    input_str = "éèëêàù"
    expected_str = "eeeeau"

    result = remove_accents(input_str)
    np.testing.assert_string_equal(result, expected_str)


@pytest.mark.parametrize("input_str, expected_str", [
    ("RE: hello world", " hello world"),
    ("re :", ""),
    ("TR: hello", " hello"),
    ("hello ", "hello "),
    ("Fwd:hello", "hello")
])
def test_remove_transfer_answer_header(input_str, expected_str):
    result = remove_transfer_answer_header(input_str)
    np.testing.assert_string_equal(result, expected_str)


@pytest.mark.parametrize("input_str, expected_str", [
    ("Bonjour, mon email : prenom.nom@hotmail.fr", "Bonjour, mon email :  flag_mail_ "),
    ("Mon numéro : 01.23.45.67.89", "Mon numéro :  flag_phone_ "),
    ("01 23 45 67 89 et 01.23.45.67.89", " flag_phone_  et  flag_phone_ "),
    ("mon numéro01 23 45 67 89", "mon numéro flag_phone_ "),
    ("le montant du contrat est du 18000$, soit 17000euros",
     "le montant du contrat est du flag_amount_ , soit flag_amount_ "),
    ("J'habite au 1 rue de la paix, Paris 75002",
     "J'habite au 1 rue de la paix, Paris flag_cp_ "),
    ("Rendez-vous le 18 décembre 2019 ou le 19/12/19 ou le 20.12.19 à 14h30",
     "Rendez-vous le  flag_date_  ou le  flag_date_ ou le  flag_date_ à  flag_time_ "),
    ("le 14/12 tu me devras 20.05 dollars", "le  flag_date_ tu me devras flag_amount_ ")
])
def test_flag_items(input_str, expected_str):
    result = flag_items(input_str)
    np.testing.assert_string_equal(result, expected_str)
